@@ -1,4 +1,6 @@
 from typing import Union, Optional, Tuple, Any, TypeVar
+from typing_extensions import Literal
+import logging
 
 import numpy as np
 import albumentations as A
@@ -11,6 +13,9 @@ from rastervision.core.data import Scene
 from rastervision.pytorch_learner.learner_config import PosInt, NonNegInt
 from rastervision.pytorch_learner.dataset.transform import (TransformType,
                                                             TF_TYPE_TO_TF_FUNC)
+from rastervision.pytorch_learner.dataset.utils import AoiSampler
+
+log = logging.getLogger(__name__)
 
 
 class AlbumentationsDataset(Dataset):
@@ -26,8 +31,15 @@ class AlbumentationsDataset(Dataset):
 
         Args:
             orig_dataset (Any): An object with a __getitem__ and __len__.
-            transform (A.BasicTransform, optional): An Albumentations
-                transform. Defaults to None.
+            transform (A.BasicTransform, optional): Albumentations
+                transform to apply to the windows. Defaults to None.
+                Each transform in Albumentations takes images of type uint8, and
+                sometimes other data types. The data type requirements can be
+                seen at https://albumentations.ai/docs/api_reference/augmentations/transforms/ # noqa
+                If there is a mismatch between the data type of imagery and the
+                transform requirements, a RasterTransformer should be set
+                on the RasterSource that converts to uint8, such as
+                MinMaxTransformer or StatsTransformer.
             transform_type (TransformType): The type of transform so that its
                 inputs and outputs can be handled correctly. Defaults to
                 TransformType.noop.
@@ -39,6 +51,7 @@ class AlbumentationsDataset(Dataset):
         self.orig_dataset = orig_dataset
         self.normalize = normalize
         self.to_pytorch = to_pytorch
+        self.transform_type = transform_type
 
         tf_func = TF_TYPE_TO_TF_FUNC[transform_type]
         self.transform = lambda inp: tf_func(inp, transform)
@@ -49,16 +62,31 @@ class AlbumentationsDataset(Dataset):
 
     def __getitem__(self, key) -> Tuple[torch.Tensor, torch.Tensor]:
         val = self.orig_dataset[key]
-        x, y = self.transform(val)
+
+        try:
+            x, y = self.transform(val)
+        except Exception as exc:
+            log.warn(
+                'Many albumentations transforms require uint8 input. Therefore, we '
+                'recommend passing a MinMaxTransformer or StatsTransformer to the '
+                'RasterSource so the input will be converted to uint8.')
+            raise exc
 
         if self.normalize and np.issubdtype(x.dtype, np.unsignedinteger):
             max_val = np.iinfo(x.dtype).max
-            x = x.astype(np.float32) / max_val
+            x = x.astype(float) / max_val
 
         if self.to_pytorch:
             # (H, W, C) --> (C, H, W)
             x = torch.from_numpy(x).permute(2, 0, 1).float()
-            y = torch.from_numpy(y)
+            if y is not None:
+                y = torch.from_numpy(y)
+
+        if y is None:
+            # Ideally, y should be None to semantically convery the absence of
+            # any label, but PyTorch's defauult collate function doesn't handle
+            # None values.
+            y = torch.tensor(np.nan)
 
         return x, y
 
@@ -88,6 +116,13 @@ class GeoDataset(AlbumentationsDataset):
             scene (Scene): A Scene object.
             transform (Optional[A.BasicTransform], optional): Albumentations
                 transform to apply to the windows. Defaults to None.
+                Each transform in Albumentations takes images of type uint8, and
+                sometimes other data types. The data type requirements can be
+                seen at https://albumentations.ai/docs/api_reference/augmentations/transforms/ # noqa
+                If there is a mismatch between the data type of imagery and the
+                transform requirements, a RasterTransformer should be set
+                on the RasterSource that converts to uint8, such as
+                MinMaxTransformer or StatsTransformer.
             transform_type (Optional[TransformType], optional): Type of
                 transform. Defaults to None.
             normalize (bool, optional): If True, x is normalized to [0, 1]
@@ -96,16 +131,6 @@ class GeoDataset(AlbumentationsDataset):
                 pytorch tensors. Defaults to True.
         """
         self.scene = scene
-        # Scene requires its raster and label sources to be activated before
-        # they can be used. However, activating and deactivating for every read
-        # could have a large overhead.
-        #
-        # Instead, we would like the source to stay activated for the entire
-        # lifespan of this dataset. To achieve this, we set stay_activated=True
-        # when activating the scene, so that once the sources are activated,
-        # they will stay activated even when the context manager exits.
-        with self.scene.activate(stay_activated=True):
-            pass
 
         super().__init__(
             orig_dataset=scene,
@@ -117,11 +142,15 @@ class GeoDataset(AlbumentationsDataset):
     def __len__(self):
         raise NotImplementedError()
 
+    @classmethod
+    def from_uris(cls, *args, **kwargs) -> 'GeoDataset':
+        raise NotImplementedError()
+
 
 T = TypeVar('T')
 
 
-def _to_tuple(x: T, n: int = 2) -> Tuple[T, T]:
+def _to_tuple(x: T, n: int = 2) -> Tuple[T, ...]:
     """Convert to n-tuple if not already an n-tuple."""
     if isinstance(x, tuple):
         if len(x) != n:
@@ -140,8 +169,11 @@ class SlidingWindowGeoDataset(GeoDataset):
                  stride: Union[PosInt, Tuple[PosInt, PosInt]],
                  padding: Optional[Union[NonNegInt, Tuple[NonNegInt,
                                                           NonNegInt]]] = None,
+                 pad_direction: Literal['both', 'start', 'end'] = 'end',
                  transform: Optional[A.BasicTransform] = None,
-                 transform_type: Optional[TransformType] = None):
+                 transform_type: Optional[TransformType] = None,
+                 normalize: bool = True,
+                 to_pytorch: bool = True):
         """Constructor.
 
         Args:
@@ -151,23 +183,47 @@ class SlidingWindowGeoDataset(GeoDataset):
                 windows.
             padding (Optional[Union[NonNegInt, Tuple[NonNegInt, NonNegInt]]]):
                 How many pixels the windows are allowed to overflow the sides
-                of the raster source. If None, padding = size.
+                of the raster source. If None, padding is set to size // 2.
                 Defaults to None.
+            pad_direction (Literal['both', 'start', 'end']): If 'end', only pad
+                ymax and xmax (bottom and right). If 'start', only pad ymin and
+                xmin (top and left). If 'both', pad all sides. Has no effect if
+                paddiong is zero. Defaults to 'end'.
             transform (Optional[A.BasicTransform], optional): Albumentations
                 transform to apply to the windows. Defaults to None.
+                Each transform in Albumentations takes images of type uint8, and
+                sometimes other data types. The data type requirements can be
+                seen at https://albumentations.ai/docs/api_reference/augmentations/transforms/ # noqa
+                If there is a mismatch between the data type of imagery and the
+                transform requirements, a RasterTransformer should be set
+                on the RasterSource that converts to uint8, such as
+                MinMaxTransformer or StatsTransformer.
             transform_type (Optional[TransformType], optional): Type of
                 transform. Defaults to None.
+            normalize (bool, optional): If True, x is normalized to [0, 1]
+                based on its data type. Defaults to True.
+            to_pytorch (bool, optional): If True, x and y are converted to
+                pytorch tensors. Defaults to True.
         """
-        super().__init__(scene, transform, transform_type)
+        super().__init__(
+            scene=scene,
+            transform=transform,
+            transform_type=transform_type,
+            normalize=normalize,
+            to_pytorch=to_pytorch)
         self.size = _to_tuple(size)
         self.stride = _to_tuple(stride)
         self.padding = padding
+        self.pad_direction = pad_direction
         self.init_windows()
 
     def init_windows(self) -> None:
         """Pre-compute windows."""
-        windows = self.scene.raster_source.get_extent().get_windows(
-            chip_sz=self.size, stride=self.stride, padding=self.padding)
+        windows = self.scene.raster_source.extent.get_windows(
+            self.size,
+            stride=self.stride,
+            padding=self.padding,
+            pad_direction=self.pad_direction)
         if len(self.scene.aoi_polygons) > 0:
             windows = Box.filter_by_aoi(windows, self.scene.aoi_polygons)
         self.windows = windows
@@ -188,17 +244,20 @@ class RandomWindowGeoDataset(GeoDataset):
 
     def __init__(self,
                  scene: Scene,
+                 out_size: Optional[Union[PosInt, Tuple[PosInt, PosInt]]],
                  size_lims: Optional[Tuple[PosInt, PosInt]] = None,
                  h_lims: Optional[Tuple[PosInt, PosInt]] = None,
                  w_lims: Optional[Tuple[PosInt, PosInt]] = None,
-                 out_size: Union[PosInt, Tuple[PosInt, PosInt]] = None,
                  padding: Optional[Union[NonNegInt, Tuple[NonNegInt,
                                                           NonNegInt]]] = None,
                  max_windows: Optional[NonNegInt] = None,
+                 max_sample_attempts: PosInt = 100,
+                 return_window: bool = False,
+                 efficient_aoi_sampling: bool = True,
                  transform: Optional[A.BasicTransform] = None,
                  transform_type: Optional[TransformType] = None,
-                 max_sample_attempts: PosInt = 100,
-                 return_window: bool = False):
+                 normalize: bool = True,
+                 to_pytorch: bool = True):
         """Constructor.
 
         Will sample square windows if size_lims is specified. Otherwise, will
@@ -207,18 +266,17 @@ class RandomWindowGeoDataset(GeoDataset):
 
         Args:
             scene (Scene): A Scene object.
+            out_size (Optional[Union[PosInt, Tuple[PosInt, PosInt]]]]): Resize
+                windows to this size before returning. This is to aid in
+                collating the windows into a batch. If None, windows are
+                returned without being normalized or converted to pytorch, and
+                will be of different sizes in successive reads.
             size_lims (Optional[Tuple[PosInt, PosInt]]): Interval from which to
                 sample window size.
             h_lims (Optional[Tuple[PosInt, PosInt]]): Interval from which to
                 sample window height.
             w_lims (Optional[Tuple[PosInt, PosInt]]): Interval from which to
                 sample window width.
-            out_size (Union[PosInt, Tuple[PosInt, PosInt]], optional): Resize
-                windows to this size before returning. This is to aid in
-                collating the windows into a batch. If None, windows are
-                returned without being normalized or converted to pytorch, and
-                will be of different sizes in successive reads.
-                Defaults to None.
             padding (Optional[Union[NonNegInt, Tuple[NonNegInt, NonNegInt]]]):
                 How many pixels the windows are allowed to overflow the sides
                 of the raster source. If None, padding = size.
@@ -228,15 +286,36 @@ class RandomWindowGeoDataset(GeoDataset):
                 np.inf. Defaults to None.
             transform (Optional[A.BasicTransform], optional): Albumentations
                 transform to apply to the windows. Defaults to None.
+                Each transform in Albumentations takes images of type uint8, and
+                sometimes other data types. The data type requirements can be
+                seen at https://albumentations.ai/docs/api_reference/augmentations/transforms/
+                If there is a mismatch between the data type of imagery and the
+                transform requirements, a RasterTransformer should be set
+                on the RasterSource that converts to uint8, such as
+                MinMaxTransformer or StatsTransformer.
             transform_type (Optional[TransformType], optional): Type of
                 transform. Defaults to None.
-            return_window (bool, optional): Make __getitem__ return the window
-                coordinates used to generate the image. Defaults to False.
             max_sample_attempts (NonNegInt, optional): Max attempts when trying
                 to find a window within the AOI of the scene. Only used if the
                 scene has aoi_polygons specified. StopIteratioin is raised if
                 this is exceeded. Defaults to 100.
-        """
+            return_window (bool, optional): Make __getitem__ return the window
+                coordinates used to generate the image. Defaults to False.
+            efficient_aoi_sampling (bool, optional): If the scene has AOIs,
+                sampling windows at random anywhere in the extent and then
+                checking if they fall within any of the AOIs can be very
+                inefficient. This flag enables the use of an alternate
+                algorithm that only samples window locations inside the AOIs.
+                Defaults to True.
+            transform (Optional[A.BasicTransform], optional): Albumentations
+                transform to apply to the windows. Defaults to None.
+            transform_type (Optional[TransformType], optional): Type of
+                transform. Defaults to None.
+            normalize (bool, optional): If True, x is normalized to [0, 1]
+                based on its data type. Defaults to True.
+            to_pytorch (bool, optional): If True, x and y are converted to
+                pytorch tensors. Defaults to True.
+        """ # noqa
         has_size_lims = size_lims is not None
         has_h_lims = h_lims is not None
         has_w_lims = w_lims is not None
@@ -246,26 +325,27 @@ class RandomWindowGeoDataset(GeoDataset):
             raise ValueError('h_lims and w_lims must both be specified')
 
         if out_size is not None:
-            normalize, to_pytorch = True, True
             out_size = _to_tuple(out_size)
             transform = self.get_resize_transform(transform, out_size)
         else:
+            log.warning(f'out_size is None, chips will not be normalized or '
+                        'converted to PyTorch Tensors.')
             normalize, to_pytorch = False, False
 
         super().__init__(
-            scene,
-            transform,
-            transform_type,
+            scene=scene,
+            transform=transform,
+            transform_type=transform_type,
             normalize=normalize,
             to_pytorch=to_pytorch)
 
         if padding is None:
             if size_lims is not None:
                 max_size = size_lims[1]
-                padding = (max_size, max_size)
+                padding = (max_size // 2, max_size // 2)
             else:
                 max_h, max_w = h_lims[1], w_lims[1]
-                padding = (max_h, max_w)
+                padding = (max_h // 2, max_w // 2)
         padding = _to_tuple(padding)
 
         if max_windows is None:
@@ -280,9 +360,19 @@ class RandomWindowGeoDataset(GeoDataset):
         self.max_sample_attempts = max_sample_attempts
 
         # include padding in the extent
-        ymin, xmin, ymax, xmax = scene.raster_source.get_extent()
+        ymin, xmin, ymax, xmax = scene.raster_source.extent
         h_padding, w_padding = self.padding
-        self.extent = (ymin, xmin, ymax + h_padding, xmax + w_padding)
+        self.extent = Box(ymin - h_padding, xmin - w_padding, ymax + h_padding,
+                          xmax + w_padding)
+
+        self.aoi_sampler = None
+        if self.scene.aoi_polygons and efficient_aoi_sampling:
+            # clip aoi polygons to the extent
+            extent_polygon = self.extent.to_shapely()
+            self.scene.aoi_polygons = [
+                p.intersection(extent_polygon) for p in self.scene.aoi_polygons
+            ]
+            self.aoi_sampler = AoiSampler(self.scene.aoi_polygons)
 
     def get_resize_transform(
             self, transform: Optional[A.BasicTransform],
@@ -323,9 +413,13 @@ class RandomWindowGeoDataset(GeoDataset):
 
     def sample_window_loc(self, h: int, w: int) -> Tuple[int, int]:
         """Randomly sample coordinates of the top left corner of the window."""
-        ymin, xmin, ymax, xmax = self.extent
-        y = torch.randint(low=ymin, high=ymax - h, size=(1, )).item()
-        x = torch.randint(low=xmin, high=xmax - w, size=(1, )).item()
+        if not self.aoi_sampler:
+            ymin, xmin, ymax, xmax = self.extent
+            y = torch.randint(low=ymin, high=ymax - h, size=(1, )).item()
+            x = torch.randint(low=xmin, high=xmax - w, size=(1, )).item()
+        else:
+            x, y = self.aoi_sampler.sample().round().T
+            x, y = int(x.item()), int(y.item())
         return x, y
 
     def _sample_window(self) -> Box:
