@@ -1,9 +1,12 @@
+from typing import Callable, Optional, Union
+from os.path import join
 from enum import Enum
-from typing import Optional, Union
-from pydantic import PositiveInt
+import logging
 
 import albumentations as A
+from torch import nn
 from torch.utils.data import Dataset
+from torchvision import models
 
 from rastervision.core.data import Scene
 from rastervision.pipeline.config import (Config, register_config, Field,
@@ -15,72 +18,95 @@ from rastervision.pytorch_learner.dataset import (
     SemanticSegmentationImageDataset,
     SemanticSegmentationSlidingWindowGeoDataset,
     SemanticSegmentationRandomWindowGeoDataset)
+from rastervision.pytorch_learner.utils import adjust_conv_channels
+
+log = logging.getLogger(__name__)
 
 
 class SemanticSegmentationDataFormat(Enum):
     default = 'default'
 
 
-def ss_data_config_upgrader(cfg_dict, version):
-    if version == 1:
+def ss_data_config_upgrader(cfg_dict: dict, version: int) -> dict:
+    if version < 2:
         cfg_dict['type_hint'] = 'semantic_segmentation_image_data'
+    elif version < 3:
+        try:
+            # removed in version 3
+            del cfg_dict['channel_display_groups']
+        except KeyError:
+            pass
+    return cfg_dict
+
+
+def ss_image_data_config_upgrader(cfg_dict: dict, version: int) -> dict:
+    if version < 3:
+        try:
+            # removed in version 3
+            del cfg_dict['img_format']
+            del cfg_dict['label_format']
+            del cfg_dict['channel_display_groups']
+        except KeyError:
+            pass
     return cfg_dict
 
 
 @register_config(
     'semantic_segmentation_data', upgrader=ss_data_config_upgrader)
 class SemanticSegmentationDataConfig(Config):
-    img_channels: PositiveInt = Field(
-        3, description='The number of channels of the training images.')
-    channel_display_groups: Optional[Union[dict, list, tuple]] = Field(
-        None,
-        description=
-        ('Groups of image channels to display together as a subplot '
-         'when plotting the data and predictions. '
-         'Can be a list or tuple of groups (e.g. [(0, 1, 2), (3,)]) or a dict '
-         'containing title-to-group mappings '
-         '(e.g. {"RGB": [0, 1, 2], "IR": [3]}), '
-         'where each group is a list or tuple of channel indices and title '
-         'is a string that will be used as the title of the subplot '
-         'for that group.'))
-
-    def update(self):
-        if self.channel_display_groups is None:
-            self.channel_display_groups = {
-                'Input': tuple(range(self.img_channels))
-            }
+    pass
 
 
-@register_config('semantic_segmentation_image_data')
+@register_config(
+    'semantic_segmentation_image_data', upgrader=ss_image_data_config_upgrader)
 class SemanticSegmentationImageDataConfig(SemanticSegmentationDataConfig,
                                           ImageDataConfig):
+    """Configure :class:`SemanticSegmentationImageDatasets <.SemanticSegmentationImageDataset>`.
+
+    This assumes the following file structure:
+
+    .. code-block:: text
+
+        <data_dir>/
+            img/
+                <img 1>.<extension>
+                <img 2>.<extension>
+                ...
+                <img N>.<extension>
+            labels/
+                <img 1>.<extension>
+                <img 2>.<extension>
+                ...
+                <img N>.<extension>
+
+    """ # noqa
     data_format: SemanticSegmentationDataFormat = (
         SemanticSegmentationDataFormat.default)
 
-    img_format: Optional[str] = Field(
-        None, description='The filetype of the training images.')
-    label_format: Optional[str] = Field(
-        None, description='The filetype of the training labels.')
-
     def update(self, *args, **kwargs):
         SemanticSegmentationDataConfig.update(self)
         ImageDataConfig.update(self, *args, **kwargs)
-        if self.img_format is None:
-            self.img_format = 'png' if self.img_channels == 3 else 'npy'
 
     def dir_to_dataset(self, data_dir: str,
                        transform: A.BasicTransform) -> Dataset:
+        if self.data_format != SemanticSegmentationDataFormat.default:
+            raise NotImplementedError()
+        img_dir = join(data_dir, 'img')
+        label_dir = join(data_dir, 'labels')
         ds = SemanticSegmentationImageDataset(
-            data_dir=data_dir,
-            img_fmt=self.img_format,
-            label_fmt=self.label_format,
-            transform=transform)
+            img_dir=img_dir, label_dir=label_dir, transform=transform)
         return ds
 
 
 @register_config('semantic_segmentation_geo_data')
 class SemanticSegmentationGeoDataConfig(SemanticSegmentationDataConfig,
                                         GeoDataConfig):
+    """Configure semantic segmentation :class:`GeoDatasets <.GeoDataset>`.
+
+    See
+    :mod:`rastervision.pytorch_learner.dataset.semantic_segmentation_dataset`.
+    """
+
     def update(self, *args, **kwargs):
         SemanticSegmentationDataConfig.update(self)
         GeoDataConfig.update(self, *args, **kwargs)
@@ -100,6 +126,7 @@ class SemanticSegmentationGeoDataConfig(SemanticSegmentationDataConfig,
                 size=opts.size,
                 stride=opts.stride,
                 padding=opts.padding,
+                pad_direction=opts.pad_direction,
                 transform=transform)
         elif opts.method == GeoDataWindowMethod.random:
             ds = SemanticSegmentationRandomWindowGeoDataset(
@@ -111,6 +138,7 @@ class SemanticSegmentationGeoDataConfig(SemanticSegmentationDataConfig,
                 padding=opts.padding,
                 max_windows=opts.max_windows,
                 max_sample_attempts=opts.max_sample_attempts,
+                efficient_aoi_sampling=opts.efficient_aoi_sampling,
                 transform=transform)
         else:
             raise NotImplementedError()
@@ -119,6 +147,8 @@ class SemanticSegmentationGeoDataConfig(SemanticSegmentationDataConfig,
 
 @register_config('semantic_segmentation_model')
 class SemanticSegmentationModelConfig(ModelConfig):
+    """Configure a semantic segmentation model."""
+
     backbone: Backbone = Field(
         Backbone.resnet50,
         description=(
@@ -133,16 +163,50 @@ class SemanticSegmentationModelConfig(ModelConfig):
                 'and resnet101.')
         return v
 
+    def build_default_model(self, num_classes: int,
+                            in_channels: int) -> nn.Module:
+        pretrained = self.pretrained
+        backbone_name = self.get_backbone_str()
+
+        model_factory_func: Callable = getattr(models.segmentation,
+                                               f'deeplabv3_{backbone_name}')
+        model = model_factory_func(
+            num_classes=num_classes,
+            pretrained_backbone=pretrained,
+            aux_loss=False)
+
+        if in_channels != 3:
+            if not backbone_name.startswith('resnet'):
+                raise ConfigError(
+                    'All TorchVision backbones do not provide the same API '
+                    'for accessing the first conv layer. '
+                    'Therefore, conv layer modification to support '
+                    'arbitrary input channels is only supported for resnet '
+                    'backbones. To use other backbones, it is recommended to '
+                    'fork the TorchVision repo, define factory functions or '
+                    'subclasses that perform the necessary modifications, and '
+                    'then use the external model functionality to import it '
+                    'into Raster Vision. See isprs_potsdam.py for an example '
+                    'of how to import external models. Alternatively, you can '
+                    'override this function.')
+            model.backbone.conv1 = adjust_conv_channels(
+                old_conv=model.backbone.conv1,
+                in_channels=in_channels,
+                pretrained=pretrained)
+        return model
+
 
 @register_config('semantic_segmentation_learner')
 class SemanticSegmentationLearnerConfig(LearnerConfig):
+    """Configure a :class:`.SemanticSegmentationLearner`."""
+
     data: Union[SemanticSegmentationImageDataConfig,
                 SemanticSegmentationGeoDataConfig]
-    model: SemanticSegmentationModelConfig
+    model: Optional[SemanticSegmentationModelConfig]
 
     def build(self,
-              tmp_dir,
-              model_path=None,
+              tmp_dir=None,
+              model_weights_path=None,
               model_def_path=None,
               loss_def_path=None,
               training=True):
@@ -151,22 +215,7 @@ class SemanticSegmentationLearnerConfig(LearnerConfig):
         return SemanticSegmentationLearner(
             self,
             tmp_dir=tmp_dir,
-            model_path=model_path,
+            model_weights_path=model_weights_path,
             model_def_path=model_def_path,
             loss_def_path=loss_def_path,
             training=training)
-
-    def validate_config(self):
-        super().validate_config()
-        self.validate_class_loss_weights()
-
-    def validate_class_loss_weights(self):
-        if self.solver.class_loss_weights is None:
-            return
-
-        num_weights = len(self.solver.class_loss_weights)
-        num_classes = len(self.data.class_names)
-        if num_weights != num_classes:
-            raise ConfigError(
-                f'class_loss_weights ({num_weights}) must be same length as '
-                f'the number of classes ({num_classes}), null class included')
